module util_win.clipboard;

private import std.outofmemory;
private import std.c.stdio;
private import win32.windows;
private import util_win.window;




/*
// clipboard format
enum {
    CF_TEXT = 1,
    CF_BITMAP,
    CF_METAFILEPICT,
    CF_SYLK,
    CF_DIF,
    CF_TIFF,
    CF_OEMTEXT,
    CF_DIB,
    CF_PALETTE,
    CF_PENDATA,
    CF_RIFF,
    CF_WAVE,
    CF_UNICODETEXT,
    CF_ENHMETAFILE,
    CF_HDROP,
    CF_LOCALE,
    CF_MAX, // = 17
    CF_OWNERDISPLAY   = 128,
    CF_DSPTEXT,
    CF_DSPBITMAP,
    CF_DSPMETAFILEPICT, // = 131
    CF_DSPENHMETAFILE = 142,
    CF_PRIVATEFIRST   = 512,
    CF_PRIVATELAST    = 767,
    CF_GDIOBJFIRST    = 768,
    CF_GDIOBJLAST     = 1023
}
*/




// set clipboard data
bool setClipboard(ubyte[] data, uint format)
{
    // open
    if (!OpenClipboard(null)) return false;
    
    try {
        // empty up
        EmptyClipboard();
        
        // allocate global memory
        HGLOBAL hGlobal = GlobalAlloc(GHND, data.length + 1);
        if (hGlobal == null) throw new Exception("setClipboard(); failed to allocate memory");

        // lock global memory
        ubyte* pHeap = cast(ubyte*)GlobalLock(hGlobal);
        if (pHeap == null) {
            GlobalFree(hGlobal);
            throw new Exception("setClipboard(); failed to lock memory");
        }

        // copy to global memory
        pHeap[0..data.length] = data;

        // unlock global memory
        GlobalUnlock(hGlobal);

        // put data
        SetClipboardData(format, hGlobal);
    }
    finally {
        // close
        CloseClipboard();
    }
    
    return true;
}




// check clipboard format
bool checkClipboard(uint format)
{
    return (IsClipboardFormatAvailable(format) != 0);
}




// get clipboard data
ubyte[] getClipboard(uint format)
{
    ubyte[] data;
    
    // check the format
    if (!IsClipboardFormatAvailable(format)) return data;

    // open
    if (!OpenClipboard(null)) return data;

    try {
        // get data
        HGLOBAL hGlobal = GetClipboardData(format);
        if (hGlobal == null) throw new Exception("getClipboard(); failed to get data");

        // lock
        ubyte* pHeap = cast(ubyte*)GlobalLock(hGlobal);
        if (pHeap == null) throw new Exception("getClipboard(); failed to lock clipboard");

        // copy from global memory
        int size = GlobalSize(hGlobal) - 1;
        data.length = size;
        data[] = pHeap[0..size];

        // unlock
        GlobalUnlock(hGlobal);
    }
    finally {
        // close
        CloseClipboard();
    }

    return data;
}




// clipboard viewer chain class
class CClipboardViewer : IWindow
{
private:
    HWND            m_hNextCBVChain = null;
    void delegate() m_dgCBVCallBack = null;

public:
    this()
    {
    }
    
    
public:
    // begin to watch clicboard 
    void begin(HWND hWnd, void delegate() dg)
    {
        formHandle(hWnd);
        m_hNextCBVChain = SetClipboardViewer(hWnd);
        m_dgCBVCallBack = dg;
    }


    // handle window message (must call from IWindow.proc())
    bool proc(HWND hWnd, uint WM, WPARAM wParam, LPARAM lParam)
    {
        if (hWnd != m_hWnd) return false;
        
        switch (WM) {
        case WM_CHANGECBCHAIN:
            if (m_hNextCBVChain == cast(HWND)wParam) m_hNextCBVChain = cast(HWND)lParam;
            if (m_hNextCBVChain) SendMessage(m_hNextCBVChain, WM_CHANGECBCHAIN, wParam, lParam);
            return true;
        case WM_DRAWCLIPBOARD:
            if (m_dgCBVCallBack) m_dgCBVCallBack();
            SendMessage(m_hNextCBVChain, WM_DRAWCLIPBOARD, 0, 0);
            return true;
        default:
            return false;
        }
    }


    // end to watch clicboard (must call)
    void end()
    {
        ChangeClipboardChain(m_hWnd, m_hNextCBVChain);
    }
}


