//------------------------------------------------------------
//  WinMain()
//------------------------------------------------------------
module util_win.winmain;


private import string = std.string;
private import win32.windows;

private import util_win.global;
private import util_win.mainwindow;


public const uint WM_CALLIDLE = WM_USER+1000;




//--------------------------------------------------------------------------------
//  Convert WinMain() to main()
//--------------------------------------------------------------------------------
extern (C) void gc_init();
extern (C) void gc_term();
extern (C) void _minit();
extern (C) void _moduleCtor();
extern (C) void _moduleDtor();

extern (Windows) 
int WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
    int result = 0;

    // Initialize something
    gc_init();
    _minit();

    try
    {
        _moduleCtor();

        // Get module name
        char exe_name[4096];
        GetModuleFileNameA(null, exe_name.ptr, 4096);
        
        // Initialize globals
        g_hInstance  = hInstance;
        g_nameModule = string.toString(exe_name.ptr);

        // Convert to char[][] arg
        char[][1] argexe = g_nameModule;

        // Call main()
        result = main(argexe ~ string.split(string.toString(lpCmdLine)));
        
        _moduleDtor();
    }

    catch (Object o)
    {
        // Catch spilled exception
        MessageBoxA(null, cast(char *)o.toString(), "Error", MB_OK | MB_ICONEXCLAMATION);
        result = 0;
    }

    // Terminate something
    gc_term();
    
    return result;
}




//------------------------------------------------------------
// main()
//------------------------------------------------------------
int main(char[][] arg)
{
    if (IMainWindow.getInstance() == null) return -1;

    main_enter(arg);
    if (!main_loop()) throw new Exception("main_loop(); Failure in main loop");
    scope(exit) main_exit();
    
    return 0;
}





//------------------------------------------------------------
// Enterance
//------------------------------------------------------------
void main_enter(char[][] arg)
{
    // Main window instance
    IMainWindow wnd = IMainWindow.getInstance();

    // Call enterInstance()
    wnd.enterInstance(arg);
 
    // Create Window
    if (!createMainWindow()) throw new Exception("main_enter(); Fail to create window");
}




//------------------------------------------------------------
// Exit
//------------------------------------------------------------
void main_exit()
{
    // Main window instance
    IMainWindow wnd = IMainWindow.getInstance();

    // Call exitInstance()
    wnd.exitInstance();
}




//------------------------------------------------------------
// Loop
//------------------------------------------------------------
bool main_loop()
{
    // Main window instance
    IMainWindow wnd = IMainWindow.getInstance();

    MSG     msg;
    bool    isHalted = false;
    
    HACCEL  hAccel = wnd.getSetting().hAccel;
    HWND    hWnd   = wnd.getHandle();

    // Message loop
    while (!isHalted)
    {
        while (PeekMessage(&msg, null, 0, 0, PM_REMOVE)) {
            // Exit loop
            if (msg.message==WM_QUIT) isHalted = true;

            // Call window procedure
            if (!TranslateAccelerator(hWnd, hAccel, &msg)) {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
            }
        }

        if (!wnd.idle()) PostMessage(hWnd, WM_CALLIDLE, 0, 0);
        
        WaitMessage();
    }
    
    return (msg.wParam == 0);
}




//--------------------------------------------------------------------------------
// Window procedure
//--------------------------------------------------------------------------------
extern (Windows) LRESULT wndProc(HWND hWnd, uint WM, WPARAM wParam, LPARAM lParam)
{
    // Main window instance
    IMainWindow wnd = IMainWindow.getInstance();

    switch (WM) {
    case WM_CREATE:
        if (wnd.getHandle() == null) wnd.formHandle(hWnd);
        break;
    case WM_DESTROY:
        PostQuitMessage(0);
        break;
    default:
        break;
    }

    // Send message to main window
    if (wnd) {
        if (wnd.proc(hWnd, WM, wParam, lParam)) return 0;
    }
    
    return DefWindowProc(hWnd, WM, wParam, lParam);
}




//--------------------------------------------------------------------------------
//  Create window
//--------------------------------------------------------------------------------
bool createMainWindow()
{
    // Main window instance
    IMainWindow     wnd     = IMainWindow.getInstance();
    WINDOW_SETTING* setting = wnd.getSetting();

    // Load icon
    if (!setting.hIcon) setting.hIcon = LoadIcon(null, IDI_APPLICATION);
    
    // 0 termination
    g_nameClass   ~= "\0";
    setting.title ~= "\0";
    
    // Set window class structure
    WNDCLASS wc;
    wc.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC | CS_DBLCLKS;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.lpfnWndProc = cast(WNDPROC)&wndProc;
    wc.hInstance = g_hInstance;
    wc.hIcon = setting.hIcon;
    wc.hCursor = LoadCursor(null, IDC_ARROW);
    wc.hbrBackground = cast(HBRUSH)(COLOR_WINDOW+1);
    wc.lpszMenuName = null;
    wc.lpszClassName = g_nameClass.ptr;

    // Register window class
    if (!RegisterClass(&wc)) return false;

    // create
    return wnd.create(HWND_DESKTOP);
}

